-- Registry for "singleton" Game Objects

-- TODO: add the functionality to the registry to delete all go.s ? 
-- TODO: what about multiple objects (not singletons): projectiles, enemies, collectibles etc. ?
-- What if the object deletes itself? Solution: delete it via registry -> unregister(boolean: delete_go ?)

require "lib.components.Player"
require "lib.components.Rock"

COMPONENTS_LIST = {Player, Component, Rock}

COMPONENT_CLASSES_REGISTRY = {}
COMPONENT_REGISTRY = {}

COMPONENT_FACTORY_ID = nil

--------------------------- INIT PART START ----------------------------------------

for _, v in ipairs(COMPONENTS_LIST) do
	if v and v.hash then 
		COMPONENT_CLASSES_REGISTRY[v.hash()] = v
	else
		print('Component is nil or hash() method is not defined!!!')
	end
end

--------------------------- INIT PART END ----------------------------------------

UNIQUE_OBJECTS_COUNTERS = {}

function unregister_components() 
	for i, v in pairs(COMPONENT_REGISTRY) do 
		if v then 
			unregister_component(i, true)
		end
	end
end

function unregister_component(component_name, with_go_delete) 
	local component = resolve_component(component_name)
	if true == with_go_delete then 
		go.delete(component.go_id)
	end
	COMPONENT_REGISTRY[component_name] = nil
end

function register_component(component_name, component) 
	if COMPONENT_REGISTRY[component_name] then 
		error("Component '" .. component_name .. "' already exists")
	end
	COMPONENT_REGISTRY[component_name] = component
end

function resolve_component(component_name) 
	local component = COMPONENT_REGISTRY[component_name]
	if not component then 
		error("Component '" .. component_name .. "' was not found in registry")
	end
	return component
end

function set_component_factory(factory_id) 
	COMPONENT_FACTORY_ID = factory_id
end

function resolve_component_class(hash_or_raw_string) 
	if not hash_or_raw_string then 
		return nil
	end
	local obj = COMPONENT_CLASSES_REGISTRY[hash_or_raw_string]
	if not obj then
		obj = COMPONENT_CLASSES_REGISTRY[hash(hash_or_raw_string)]
	end
	return obj
end

function create_component(hash_or_raw_string_of_component_type, component_name, opts)
	-- TODO: need to pass pos somehow on the go creation phase
	local is_singleton = true
	if opts.is_singleton ~= nil then 
		is_singleton = opts.is_singleton
	end
	if is_singleton ~= true then 
		local id_suffix_counter = UNIQUE_OBJECTS_COUNTERS[component_name]
		if id_suffix_counter then 
			id_suffix_counter = id_suffix_counter + 1
		else
			id_suffix_counter = 0
		end
		UNIQUE_OBJECTS_COUNTERS[component_name] = id_suffix_counter
		component_name = component_name .. '_' .. id_suffix_counter
	end
	local go_obj = factory.create(COMPONENT_FACTORY_ID, opts.pos, nil, {acquire_input = opts.acquire_input})
	opts.id = go_obj
	opts.name = component_name
	local obj = resolve_component_class(hash_or_raw_string_of_component_type)(opts)
	register_component(component_name, obj)
	msg.post(go_obj, "set_delegate", {component_name = component_name})
	obj:set_create_component_function(create_component)
	obj:set_unregister_component_function(unregister_component)
	return obj
end